'Program Bootloadera (monitora) dla ATmega8 umoliwiajcy adowanie
'programu do pamici Flash przez interfejs RS232 z prdkoci 19200
'bodw

'Bity fuse naley ustawi, by na Bootloader byo przeznaczonych 128
'bajtw
'Pocztek adresu programu Bootloadera: $F80

'Fuse bity:       Wielko:  Miejsce:    Bootloader:  Koniec:  Pocztek:
'BOOTSZ1 BOOTSZ0
'  1        1      128      $000 - $F7F  $F80 - $FFF    $F7F     $F80

'Bootloader pobiera plik w standardzie Intel HEX
'Przykad pliku HEX:
':Size Address Type Data Bytes      Checksum
':10 00 00 00 26 C0 B3 C0 B3 C0 B3 C0 B3 C0 C5 C0 C5 C0 D0 C0  A4
':10 00 10 00 DB C0 E4 C0 ED C0 30 31 32 33 34 35 36 37 38 39  E7
'...
'...
':10 05 30 00 55 DF 08 95 57 E5 57 DF 52 E5 55 DF 54 E5 53 DF  A2
':10 05 40 00 5C E2 51 DF 53 2F 52 95 49 DF 53 2F 47 DF 01 D0  33
':0B 05 50 00 08 95 5D E0 48 DF 5A E0 46 DF 08 95 A2    DC
':00 00 00 01                                                     FF

$regfile = "m8def.dat"                                      'informuje kompilator o pliku
       'dyrektyw mikrokontrolera
$crystal = 8000000                                          'informuje kompilator
       'o czstotliwoci oscylatora
       'taktujcego mikrokontroler
Const Ramend_lo = $5f                                       'ustawienie adresu koca pamici
       'RAM
Const Ramend_hi = $04
Const Pagesize = 32                                         'rozmiar strony pamici Flash
       'w sowach
'definicje zmiennych
!.def Tmp_reg = R16                                         'rejestr tymczasowy
!.def Hex_reg = R17                                         'rejestr przeliczeniowy HEX
!.def Ser_reg = R18                                         'rejestr bdcy buforem
       'szeregowym
!.def SPM_reg = R19                                         'tymczasowy rejestr dla rozkazu
       'SPM
!.def Rec_size = R20                                        'numer bajtu danych w linii
       'pliku HEX
!.def Chk_sum = R21                                         'przechowuje sum kontrolna

'pocztek programu umoliwiajcego wywoanie Bootloadera
Config Pinb.1 = Input                                       'linia PB1 jako wejciowa

S1 Alias Pinb.1                                             'przypisanie nazwie Pinb.1
       'nazwy S1
Set Portb.1                                                 'doczenie do linii PB1
       'rezystora podcigajcego
Do                                                          'nieskoczona ptla Do...Loop
 If S1 = 0 Then                                             'jeli naciniety przycisk S1, to
  Waitms 25                                                 'opnienie 25 ms dla eliminacji
       'drga stykw
  If S1 = 0 Then                                            'sprawdzenie ponownie, czy S1
       'jest naciniety, jeli tak, to
   JMP $F80                                                 'skok do programu Bootloadera
  End If                                                    'koniec warunkw
 End If
Loop                                                        'koniec nieskoczonej ptli
End                                                         'koniec programu

'*********************************************************************

'Pocztek Bootloadera napisanego w asemblerze

$boot = $f7f                                                'ustawienie wektora Boot na
       'adres F7F, od tego adresu
       'zostanie zaadowany do pamici
       'program Bootloadera
 NOP                                                        'instrukcji NOP

Disable Interrupts                                          'wyczenie globalnych przerwa

_bootloader_start:                                          'pocztek Bootloadera
 LDI   Tmp_reg, Ramend_Hi                                   'wpisanie do Tmp_reg grnej
       'wartoci adresu pamici SRAM
 !OUT   SPH, Tmp_reg                                        'ustawienie bardziej znaczcego
       'bajtu wskanika stosu
 LDI   Tmp_reg, Ramend_Lo                                   'wpisanie do Tmp_reg dolnej
       'wartoci adresu pamici SRAM
 !OUT   SPL, Tmp_reg                                        'ustawienie mniej znaczcego
       'bajtu wskanika stosu
 LDI   Tmp_reg, $00                                         'wpisanie do Tmp_reg ustawienia
 'usartu
 !OUT   UCSRA, Tmp_reg                                      'ustawienie USARTu
 LDI   Tmp_reg, $18                                         'wpisanie do Tmp_reg nastpnych
       'ustawie USARTu
 !OUT   UCSRB, Tmp_reg                                      'ustawienie USARTu
 LDI   Tmp_reg, $86                                         'wpisanie do Tmp_reg nastpnych
       'ustawie USARTu
 !OUT   UCSRC, Tmp_reg                                      'ustawienie USARTu
 LDI   Tmp_reg, $00                                         'wpisanie do Tmp_reg bardziej
       'znaczcego bajtu, ktry okrela
       'prdko
 !OUT   UBRRH, Tmp_reg                                      'ustawienie USARTu
 LDI   Tmp_reg, 25                                          'wpisanie do Tmp_reg mniej
       'znaczcego bajtu, ktry okrela
       'prdko
 !out   UBRRL,tmp_reg                                       'ustawienie USARTu
 CLT                                                        'zerowanie flagi T, uytej do
       'wskazania koca pliku HEX
_send_boot_msg:                                             'wysya znak B oznaczajcy start
       'Bootloadera
 LDI   Ser_reg, Asc("B")                                    'wpisanie znaku B do Ser_reg
 RCALL _send_ser                                            'wysanie znaku B przez RS232

_read_lines:                                                'odczytuje linie z szeregowego
       'interfejsu RS232 do pamici
       'SRAM
 RCALL _receive_hex_line                                    'otrzymanie pojedynczej linii
       'zapisanej do pamici SRAM
_parse_line:                                                'dekoduje wybrana linie HEX
 LDI   XH, $01                                              'pocztek linii, bardziej
       'znaczcy bajt, od $0100
 LDI   XL, $00                                              'pocztek linii, mniej znaczcy
       'bajt
 CLR   Chk_sum                                              'wyzerowanie rejestru sumy
       'kontrolnej dla tej linii
_read_header:                                               'odczytuje dane
 LD    Tmp_reg, X+                                          'pierwszym otrzymanym znakiem
       'powinien by znak :
 CPI   Tmp_reg, Asc(":")                                    'porwnanie czy otrzymano : i
       'ustawienie flagi bdu
 BREQ  _header_ok                                           'odczytanie kolejnego rekordu

_header_err:                                                'jeli nie otrzymano :
 LDI   Ser_reg, Asc("!")                                    'wpisanie znaku bdu !
 RCALL _send_ser                                            'wysanie znaku bdu

_header_ok:                                                 'odczyt nastpnego rekordu
_read_record_size:                                          'odczytanie danych wybranej
       'linii
 RCALL _char_to_byte                                        'wywoanie podprogramu konwersji
       'znaku na bajt
 MOV   Rec_size, Hex_reg                                    'zapis numeru bajtu w tej linii
 TST   Rec_size                                             'test, czy rekord ma warto 0
       'dla tej linii
 BRNE _read_address                                         'jeli nie, to nie koniec linii
       'i kontynuacja odczytu
 !SET                                                       'ustawienie flagi T kiedy
       'wskazywana jest ostatnia strona
 RJMP _write_current_page                                   'koniec linii, zapisanie
       'aktualnej strony i wyjcie do
       'gwnego programu
_read_address:                                              'odczytanie bajtu bardziej
       'i mniej znaczcego adresu
       'i zapis do ZH/ZL
 RCALL _char_to_byte                                        'wywoanie podprogram konwersji
       'znaku na bajt
 MOV   ZH, Hex_reg                                          'wpisanie do ZH strony z bajtem
       'bardziej znaczcym adresu
 RCALL _char_to_byte                                        'wywoanie podprogramu konwersji
       'znaku na bajt
 MOV   ZL, Hex_reg                                          'wpisanie do ZH strony z bajtem
       'mniej znaczcym adresu
_read_record_type:                                          'odczyt typu rekordu linii
 RCALL _char_to_byte                                        'wywoanie podprogramu konwersji
       'znaku na bajt
_read_data_pairs:                                           'odczyt pozostaych bajtw
       'danych
 RCALL _char_to_byte                                        'wywoanie podprogramu konwersji
       'znaku na bajt
 MOV   R0, Hex_reg                                          'zapisanie w R0 mniej znaczcego
       'bajtu strony zapisanej
       'w buforze
 RCALL _char_to_byte                                        'wywoanie podprogramu konwersji
       'znaku na bajt
 MOV   R1, Hex_reg                                          'zapisanie w R1 bardziej
       'znaczcego bajtu strony
       'zapisanej w buforze
_store_word:                                                'zapisanie aktualnych R1/R0
       'sowa w stronie bufora pod
       'odpowiednim adresem
 LDI   Spm_reg, 1                                           'wczenie SPM (Enable (SPMEN))
 RCALL _exec_spm                                            'wykonanie aktualnego SPM

_check_byte_count:                                          'sprawdzenie, czy byo to
       'ostatnie sowo linii
 SUBI  Rec_size, 2                                          'zmniejszenie Rec_Size o dwa
       'bajty
 BREQ _read_checksum                                        'po zapisie danych czytana jest
       'suma kontrolna
 SUBI  ZL, $FE                                              'gdy dane jeszcze nie zapisane,
       'to inkrementowane dwa mniej
       'znaczce bajty adresu
 RJMP  _read_data_pairs                                     'odczytanie nastpnych dwch
       'znakw

_read_checksum:                                             'sprawdza sum kontrolna
       'nastpnej linii
 RCALL _char_to_byte                                        'wywoanie podprogramu konwersji
       'znaku na bajt
 BREQ  _checksum_ok                                         'jeli 0, to suma kontrolna
       'prawidowa

_checksum_err:                                              'jeli suma kontrolna
       'nieprawidowa lub inny
       'znaleziony bd, to
 LDI   Ser_reg, Asc("!")                                    'wpisanie znaku ! bdcego flag
       'bdu
 RCALL _send_ser                                            'wysanie zaadowanego znaku

_checksum_ok:
_chk_page_full:                                             'sprawdza, czy strona jest pena
 MOV   Tmp_reg, ZL                                          'wpisanie mniej znaczcego bajtu
 'adresu Do Tmp_reg
 ANDI  Tmp_reg, ((Pagesize - 1 )*2)                         'iloczyn logiczny z wielkoci
       'strony
 CPI   Tmp_reg, ((Pagesize - 1 )*2)                         'porwnanie z wielkoci strony
       'dla tego adresu
 BRNE  _read_lines                                          'strona bufora nie jest pena,
       'odczyt nastpnej linii
_write_current_page:                                        'zapis aktualnej strony
_erase_page:                                                'wyczyszczenie strony
 LDI   Spm_reg, 3                                           'odblokowanie SPM
 RCALL _exec_spm                                            'wykonanie SPM

_write_page:                                                'zapisuje stron
 ANDI  ZL, $C0                                              'zapewnienie, e bity Z5...Z0 s
       '0
 LDI   Spm_reg, 5                                           'zapis strony i odblokowanie SPM
 RCALL _exec_spm                                            'wykonanie SPM

_enable_page:                                               'wczenie strony dla sekcji
       'Read-While-Write
 RCALL _wait_spm                                            'sprawdzenie, czy zapis kompletny
 LDI   Spm_reg, 11                                          'ustawienie tylko RWWSRE i SPMEN
 RCALL _exec_spm                                            'wykonanie SPM

_check_end_of_file:                                         'sprawdzenie, czy to nie jest
       'koniec pliku
 BRTC  _read_lines                                          'jeli nie jest to ostatnia
       'strona, to kontynuacja odczytu
       'linii

_exit_bootloader:                                           'wyjcie z programu Bootloader
 JMP $0000                                                  'skok do wektora Reset

'********************************************************************
_send_ser:                                                  'wysya znak przez RS232
 SBIS  UCSRA,UDRE                                           'sprawdza, czy rejestr nadawczy
       'USARTu jest pusty
 RJMP  _send_ser                                            'jeli nie, to czekaj
 !OUT   UDR,ser_reg                                         'wysanie znaku
 RET                                                        'koniec podprogramu

'********************************************************************
'pobiera jedna linie przez RS232 i zapisuje j od pocztku pamici
'SRAM
_receive_hex_line:
 LDI   XH, $01                                              'ustawienie wskanika do SRAM na
       'adres $0100
 LDI   XL, $00                                              'powyej s wszystkie rejestry

 LDI   Ser_reg, Asc("?")                                    'aduje znak ? dla dania
       'nastpnego znaku
 RCALL _send_ser                                            'wywoanie podprogramu wysania
       'znaku

_receive_hex_line_char:                                     'pobiera znak z USARTu i dodaje
       'go do bufora
 SBIS  UCSRA, RXC                                           'sprawdza, czy otrzymano znak
       'z USARTu
 RJMP  _receive_hex_line_char                               'jeli nie, to ponowne
       'sprawdzenie
 IN    Tmp_reg, UDR                                         'zapisanie otrzymanego znaku do
 'tmp_reg
' MOV   Ser_reg, Tmp_reg    'echo dla kontroli odebranych
       'znakw
' RCALL _send_ser     'wywoanie podprogramu wysania
       'znaku

 CPI   Tmp_reg, 13                                          'porwnanie otrzymanego znaku
       'z CR,
 BREQ  _receive_hex_line_end                                'jeli taki sam, to koniec
       'linii,
 ST    X+, Tmp_reg                                          'w przeciwnym razie zwikszenie
       'wskanika buforu
 RJMP  _receive_hex_line_char                               'i odczytanie nastpnego znaku

_receive_hex_line_end:                                      'linia wykonywana, gdy koniec
       'linii
 RET                                                        'powrt z tego podprogramu

'********************************************************************
'pobiera dwa znaki z bufora, dodaje sum kontrolna i zwraca rezultat
'zamiany znakw na bajty w Hex_reg
_char_to_byte:
 LD    Hex_reg, X+                                          'wpisanie znaku do Hex_reg
       'i zwikszenie X
 SUBI  Hex_reg, $41                                         'odjcie znaku ASCII $41, (A)
 BRCC  _char_to_byte1                                       'skok, gdy wartoci s wiksze
       'od $41,
 SUBI  Hex_reg, $F9                                         'gdy nie wiksze, to odejcie
       '$F9
_char_to_byte1:
 SUBI  Hex_reg, $F6                                         'odjcie $F6
 LSL   Hex_reg                                              'przesunicie bitw w Hex_reg o
 LSL   Hex_reg                                              '4 pozycje w lewo
 LSL   Hex_reg
 LSL   Hex_reg
 LD    Tmp_reg, X+                                          'pobranie nastpnego znaku
 SUBI  Tmp_reg, $41                                         'odjcie znaku ASCII $41, (A)
 BRCC  _char_to_byte2                                       'skok, gdy wartoci s wiksze
       'od $41,
 SUBI  Tmp_reg, $F9                                         'gdy nie wiksze, to odjcie $F9
_char_to_byte2:
 SUBI  Tmp_reg, $F6                                         'odjcie $F6

 ADD   Hex_reg, Tmp_reg                                     'dodanie Tmp_reg do Hex_reg
 ADD   Chk_sum, Hex_reg                                     'dodanie do sumy kontrolnej
       'Hex_reg
 RET                                                        'powrt z podprogramu

'********************************************************************
_exec_spm:                                                  'wykonuje rozkaz SPM
 !OUT   SPMCR, Spm_reg                                      'wysanie do SPM wartoci
       'Spm_reg
 SPM                                                        'rozkaz SPM

_wait_spm:                                                  'sprawdzenie, czy zapis do
       'pamici Flash kompletny
 IN    Spm_reg, SPMCR                                       'pobranie do Spm_reg SPM
 SBRC  Spm_reg, SPMEN                                       'sprawdzenie, czy SPM Enable
       'flaga jest wyzerowana
 RJMP  _wait_spm                                            'jeli nie, to czekanie a flaga
       'bdzie wyzerowana
 RET                                                        'flaga wyzerowana, powrt
       'podprogramu
'********************************************************************
